import argparse
import datetime
import json
import sys
from pathlib import Path

import numpy as np
import yaml
from marine.logger import getLogger
from marine.utils.g2p_util import pron2mora
from tqdm import tqdm

logger = None

UNUSED_SYMBOL_REMOVER = str.maketrans("", "", "^$[?")
ACCENT_NUCLEUS_SYMBOL = "]"
ACCENT_PHRASE_BOUNDARY_SYMBOL = "#"
INTONATION_PHRASE_BOUNDARY_SYMBOL = "_"
INTONATION_PHRASE_BOUNDARY_PUNCTUATION = ","


def get_parser():
    parser = argparse.ArgumentParser(
        description="Convert Special format txt format data to json file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("in_path", type=Path, help="Input path or directory")
    parser.add_argument("out_dir", type=Path, help="Output directory")
    parser.add_argument(
        "--accent_status_seq_level",
        "-s",
        type=str,
        choices=["ap", "mora"],
        default="ap",
        help="Sequence level for accent status label",
    )
    parser.add_argument(
        "--accent_status_represent_mode",
        "-m",
        type=str,
        choices=["binary", "high_low"],
        default="binary",
        help="""Representation mode for accent status label
        (this option will be ignored when --accent-status-seq-level is chosen as 'ap')""",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        type=int,
        default=50,
        help="Logging level",
    )
    return parser


def alignment_feature(features, target_feature, ignore_features):
    # filtering
    for ignore_feature in ignore_features:
        features = features[features != ignore_feature]

    # detect
    _features = features == target_feature

    # move & pad
    mask = ~(np.concatenate([_features[1:], [False]]))
    _features = _features[mask]

    if target_feature in [
        ACCENT_PHRASE_BOUNDARY_SYMBOL,
        INTONATION_PHRASE_BOUNDARY_SYMBOL,
    ]:
        _features = np.concatenate([[False], _features[:-1]])

    return _features


def convert_mask_seq_to_int_seq(feature):
    return np.where(feature, 1, 0)


def merge_ip_ap_boundary(accent_phrase_boundaries, intonation_phrase_boundaries):
    assert len(accent_phrase_boundaries) == len(intonation_phrase_boundaries)

    return accent_phrase_boundaries + intonation_phrase_boundaries


def binary_accent_to_ap_accent(
    binary_accents,
    accent_phrase_boundaries,
    accent_label=1,
    accnet_phrase_label=1,
):
    assert len(binary_accents) == len(accent_phrase_boundaries)

    accent_phrase_boundary_indexes = np.where(
        accent_phrase_boundaries == accnet_phrase_label
    )[0]
    splitted_binary_accents = np.split(binary_accents, accent_phrase_boundary_indexes)

    ap_accent_labbels = []

    for binary_accent in splitted_binary_accents:
        accent_indexs = np.where(binary_accent == accent_label)[0]

        if len(accent_indexs) >= 1:
            acc = accent_indexs[0] + 1  # zero pad
        else:
            acc = 0

        ap_accent_labbels.append(acc)

    return ap_accent_labbels


def binary_accent_to_high_low_accent(
    moras,
    binary_accents,
    accent_phrase_boundaries,
    accent_label=1,
    accnet_phrase_label=1,
    accent_status_represent_mode="high_low",
):
    assert len(moras) == len(binary_accents) == len(accent_phrase_boundaries)

    accent_phrase_boundary_indexes = np.where(
        accent_phrase_boundaries == accnet_phrase_label
    )[0]
    splitted_moras = np.split(moras, accent_phrase_boundary_indexes)
    splitted_binary_accents = np.split(binary_accents, accent_phrase_boundary_indexes)

    high_low_accent_labels = []

    for mora, binary_accent in zip(splitted_moras, splitted_binary_accents):
        accent_indexs = np.where(binary_accent == accent_label)[0]

        if len(accent_indexs) >= 1:
            _acc = accent_indexs[0] + 1  # zero pad
            _, acc = pron2mora(mora, int(_acc), accent_status_represent_mode)
            if accent_status_represent_mode == "high_low" and mora[int(_acc)] == "ー":
                acc[int(_acc)] = 0
        else:
            _, acc = pron2mora(mora, 0, accent_status_represent_mode)

        high_low_accent_labels += acc

    return high_low_accent_labels


def convert_to_srt_feature(features, splitter=","):
    if isinstance(features, np.ndarray):
        features = features.tolist()
    elif not isinstance(features, list):
        raise TypeError("Wrong type of feature")

    return splitter.join([str(value) for value in features])


def parse_jsut_annotation(
    annotation, accent_status_seq_level, accent_status_represent_mode
):
    features = {}

    # preprocessing: remove unused symbols
    annotation = annotation.translate(UNUSED_SYMBOL_REMOVER)
    # preprocessing: replace ヲ -> オ
    annotation = annotation.replace("ヲ", "オ")
    # preprocessing: parse as sequence
    mora_based_annotation = np.array(pron2mora(annotation))

    # filtering symbol
    moras = mora_based_annotation[
        (mora_based_annotation != ACCENT_NUCLEUS_SYMBOL)
        & (mora_based_annotation != ACCENT_PHRASE_BOUNDARY_SYMBOL)
        & (mora_based_annotation != INTONATION_PHRASE_BOUNDARY_SYMBOL)
    ]

    assert len(moras) > 0, "Empty annotation"

    binary_accents = alignment_feature(
        mora_based_annotation,
        ACCENT_NUCLEUS_SYMBOL,
        ignore_features=ACCENT_PHRASE_BOUNDARY_SYMBOL
        + INTONATION_PHRASE_BOUNDARY_SYMBOL,
    )
    accent_phrase_boundaries = alignment_feature(
        mora_based_annotation,
        ACCENT_PHRASE_BOUNDARY_SYMBOL,
        ignore_features=ACCENT_NUCLEUS_SYMBOL + INTONATION_PHRASE_BOUNDARY_SYMBOL,
    )
    intonation_phrase_boundaries = alignment_feature(
        mora_based_annotation,
        INTONATION_PHRASE_BOUNDARY_SYMBOL,
        ignore_features=ACCENT_NUCLEUS_SYMBOL + ACCENT_PHRASE_BOUNDARY_SYMBOL,
    )

    assert (
        len(moras)
        == len(binary_accents)
        == len(accent_phrase_boundaries)
        == len(intonation_phrase_boundaries)
    ), "{} != {} != {} != {}".format(
        len(moras),
        len(binary_accents),
        len(accent_phrase_boundaries),
        len(intonation_phrase_boundaries),
    )

    accents = convert_mask_seq_to_int_seq(binary_accents)
    accent_phrase_boundaries = convert_mask_seq_to_int_seq(accent_phrase_boundaries)
    intonation_phrase_boundaries = convert_mask_seq_to_int_seq(
        intonation_phrase_boundaries
    )

    accent_phrase_boundaries = merge_ip_ap_boundary(
        accent_phrase_boundaries, intonation_phrase_boundaries
    )

    if accent_status_seq_level == "ap":
        accents = binary_accent_to_ap_accent(accents, accent_phrase_boundaries)
    elif accent_status_seq_level == "mora" and accent_status_represent_mode != "binary":
        accents = binary_accent_to_high_low_accent(
            moras,
            accents,
            accent_phrase_boundaries,
            accent_status_represent_mode=accent_status_represent_mode,
        )

    features["pron"] = convert_to_srt_feature(moras, splitter="")
    features["accent_status"] = convert_to_srt_feature(accents)
    features["accent_phrase_boundary"] = convert_to_srt_feature(
        accent_phrase_boundaries
    )
    features["intonation_phrase_boundary"] = convert_to_srt_feature(
        intonation_phrase_boundaries
    )

    return features


def load_jsut_corpus(
    jsut_corpus_dir, accent_status_seq_level, accent_status_represent_mode
):
    text_yaml_path = jsut_corpus_dir / "text_kana" / "basic5000.yaml"
    annotation_yaml_path = jsut_corpus_dir / "e2e_symbol" / "katakana.yaml"

    scripts = []
    texts = {}
    annotations = {}

    with open(text_yaml_path, "r") as file:
        texts = yaml.safe_load(file)

    with open(annotation_yaml_path, "r") as file:
        annotations = yaml.safe_load(file)

    assert texts.keys() == annotations.keys(), "Not matched text and annotations"

    for script_id in tqdm(texts.keys(), "Parse anntoations"):
        features = {}

        surface = texts[script_id]["text_level0"]
        annotation = annotations[script_id]
        feature = parse_jsut_annotation(
            annotation, accent_status_seq_level, accent_status_represent_mode
        )

        features["script_id"] = script_id
        features["surface"] = surface
        features.update(feature)

        scripts.append(features)

    logger.info(f"Loaded {len(scripts)} scripts")

    return scripts


def entry(argv=sys.argv):
    global logger

    args = get_parser().parse_args(argv[1:])
    logger = getLogger(args.verbose)
    logger.debug(f"Loaded parameters: {args}")

    scripts = load_jsut_corpus(
        args.in_path, args.accent_status_seq_level, args.accent_status_represent_mode
    )

    if not args.out_dir.exists():
        args.out_dir.mkdir(parents=True)

    today = datetime.date.today().strftime("%y%m%d")
    with open(args.out_dir / f"just_corpus_{today}.json", "w") as file:
        json.dump(scripts, file, ensure_ascii=False, indent=4, separators=(",", ": "))


if __name__ == "__main__":
    sys.exit(entry())



import json
import random
import time
from pathlib import Path
from shutil import copyfile

import hydra
import torch
from hydra.utils import to_absolute_path
from marine.bin.test import test_model
from marine.data.feature.feature_set import FeatureSet
from marine.data.util import load_dataset
from marine.logger import getLogger
from marine.models import (
    AttentionBasedLSTMDecoder,
    CRFDecoder,
    LinearDecoder,
    init_model,
)
from marine.utils.metrics import MultiTaskMetrics
from marine.utils.util import (
    init_seed,
    log_scores,
    pack_inputs,
    pack_outputs,
    pad_incomplete_accent_logits,
    plot_batch_attention,
)
from omegaconf import DictConfig, OmegaConf
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

logger = None


def save_checkpoint(
    config, checkpoint_dir, model, optimizer, scheduler, epoch, is_best
):
    if model is None:
        return

    # save config file
    if epoch == 0:
        if not config.train.save_vocab_path:
            config.model.vocab_path = None

        config_path = checkpoint_dir / "config.yaml"
        logger.info(f"Save config: {config_path}")

        OmegaConf.save(config, config_path)

    is_interval = epoch > 0 and (epoch + 1) % config.train.checkpoint_interval == 0

    # save checkpoint when epoch is interval
    if is_interval or is_best:
        states = {
            "state_dict": model.state_dict(),
            "epoch": epoch,
            "optimizer": optimizer.state_dict()
            if config.train.save_optimizer_state and optimizer is not None
            else None,
            "scheduler": scheduler.state_dict()
            if config.train.save_optimizer_state and scheduler is not None
            else None,
        }

        if is_interval:
            interval_checkpoint_path = checkpoint_dir / f"epoch_{epoch:05d}.pth"
            lastest_path = checkpoint_dir / "latest.pth"
            logger.info(f"Save interval checkpoint: {interval_checkpoint_path}")
        else:
            interval_checkpoint_path = None
            lastest_path = None

        if is_best:
            best_checkpoint_path = checkpoint_dir / "best.pth"
            logger.info(f"Save best checkpoint: {best_checkpoint_path}")
        else:
            best_checkpoint_path = None

        if is_interval and is_best:
            torch.save(states, interval_checkpoint_path)
            copyfile(interval_checkpoint_path, best_checkpoint_path)
            copyfile(interval_checkpoint_path, lastest_path)

        elif is_interval:
            torch.save(states, interval_checkpoint_path)
            copyfile(interval_checkpoint_path, lastest_path)

        else:
            torch.save(states, best_checkpoint_path)


def train_model(
    model,
    criterions,
    optimizer,
    scheduler,
    dataloader,
    tasks,
    tensorboard_writer,
    checkpoint_dir,
    config,
    feature_set,
    num_epochs=10,
    device="cpu",
):
    phases = ["train", "val"]
    min_val_loss = [100.0] * len(tasks)
    fig_logging_targets = random.choices(range(config.data.batch_size), k=10)

    if "accent_status" in tasks:
        has_att_based_model = isinstance(
            model.decoders["accent_status"], AttentionBasedLSTMDecoder
        )
    else:
        has_att_based_model = False

    for epoch in range(num_epochs):
        logger.info("-" * 10)
        logger.info("Epoch {}/{}".format(epoch, num_epochs - 1))
        logger.info("-" * 10)

        since = time.time()

        for phase in phases:
            is_train = phase == "train"

            if is_train:
                model.train()
            else:
                model.eval()

            # for logging
            running_loss = {task: 0.0 for task in tasks}
            metrics = MultiTaskMetrics(
                phase,
                config.data.output_sizes,
                accent_represent_mode=config.data.represent_mode,
                require_ap_level_f1_score=has_att_based_model,
                device=device,
            )

            for batch_index, (inputs, outputs, _, script_ids) in enumerate(
                tqdm(dataloader[phase], desc=f"{phase}: ", leave=False)
            ):
                # pack inputs to device
                inputs = pack_inputs(inputs, config.data.input_keys, device)
                outputs = pack_outputs(outputs, device)

                # total loss of the tasks on a single batch
                batch_loss = 0.0
                prev_decoder_output = {}

                optimizer.zero_grad()

                for index, task in enumerate(tasks):
                    output, output_mask = outputs[task]["label"], outputs[task]["mask"]

                    # predict
                    with torch.set_grad_enabled(is_train):
                        decoder_outputs = model(task, **inputs)

                        if isinstance(model.decoders[task], CRFDecoder):
                            linear_logits, crf_logits = decoder_outputs
                            loss = criterions[task](linear_logits, output, output_mask)
                            logits = crf_logits
                        elif isinstance(model.decoders[task], LinearDecoder):
                            logits = decoder_outputs
                            loss = criterions[task](logits, output, output_mask)
                        elif isinstance(
                            model.decoders[task], AttentionBasedLSTMDecoder
                        ):
                            logits, attentions, ap_lengths = decoder_outputs
                            # plot attention when first batch on eval
                            if (
                                not is_train
                                and tensorboard_writer
                                and batch_index == 0
                                and task == "accent_status"
                            ):
                                plot_batch_attention(
                                    inputs,
                                    logits,
                                    ap_lengths,
                                    attentions,
                                    feature_set,
                                    plot_targets=fig_logging_targets,
                                    tensorboard_writer=tensorboard_writer,
                                    phase=phase,
                                    epoch=epoch,
                                    script_ids=script_ids,
                                )

                            # resize logits
                            logits = pad_incomplete_accent_logits(logits, output_mask)
                            loss = criterions[task](logits, output, output_mask)

                        batch_loss += loss
                        running_loss[task] += loss

                        # logits: (B, T, dim) -> (B, T)
                        predicts = torch.argmax(logits, dim=2)

                        # Update metrics
                        # for ap-based seq
                        if task == "accent_status" and has_att_based_model:
                            metrics.update(
                                task,
                                predicts,
                                output,
                                output_mask,
                                predicted_accent_phrase_boundaries=inputs[
                                    "prev_decoder_outputs"
                                ]["accent_phrase_boundary"],
                                target_accent_phrase_boundaries=outputs[
                                    "accent_phrase_boundary"
                                ]["label"],
                                mora_seq_masks=outputs["accent_phrase_boundary"][
                                    "mask"
                                ],
                            )

                        # for mora-based seq
                        else:
                            metrics.update(task, predicts, output, output_mask)

                        # cascade prev outputs
                        if index < len(tasks) - 1:
                            prev_decoder_output[task] = predicts

                            if is_train:
                                # Use golden label as teacher forcing
                                inputs["prev_decoder_outputs"][task] = output

                                if (
                                    index + 1 < len(tasks)
                                    and tasks[index + 1] == "accent_status"
                                    and has_att_based_model
                                ):
                                    inputs["decoder_targets"] = outputs[
                                        "accent_status"
                                    ]["label"]
                            else:
                                _output = prev_decoder_output[task]
                                inputs["prev_decoder_outputs"][task] = _output
                                inputs["decoder_targets"] = None

                if is_train:
                    batch_loss.backward()
                    optimizer.step()

            epoch_loss = {
                task: running_loss[task] / len(dataloader[phase]) for task in tasks
            }

            # Logging scores
            log_scores(
                phase,
                epoch,
                tasks,
                metrics,
                loss=epoch_loss,
                tensorboard_writer=tensorboard_writer,
            )

            if not is_train:
                current_val_losses = list(epoch_loss.values())

                # update schedule with validation loss
                scheduler.step(sum(current_val_losses))
                epoch_lr = optimizer.param_groups[0]["lr"]
                tensorboard_writer.add_scalar(f"{phase}_learning_rate", epoch_lr, epoch)

                # Save checkpoints
                is_best = [loss.item() for loss in current_val_losses] < min_val_loss
                min_val_loss = current_val_losses

                save_checkpoint(
                    config,
                    checkpoint_dir,
                    model,
                    optimizer,
                    scheduler,
                    epoch,
                    is_best,
                )

        time_elapsed = time.time() - since
        logger.info(f"complete in {time_elapsed // 60:.0f}m {time_elapsed % 60:.0f}s")


@hydra.main(config_path="conf/train", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.train.verbose)
    logger.info(OmegaConf.to_yaml(config))

    init_seed(config.train.seed)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    num_epochs = config.train.num_epochs

    dataloader = load_dataset(config)
    tasks = config.data.output_keys

    checkpoint_dir = (
        Path(to_absolute_path(config.train.out_dir)) / config.train.model_name
    )

    if not checkpoint_dir.exists():
        logger.info(f"Created checkpoint dir at {checkpoint_dir}")
        checkpoint_dir.mkdir(parents=True)

    # Setup tensorboard summary writer
    if config.train.tensorboard_event_path is None:
        tensorboard_event_path = f"tensorboard/{checkpoint_dir.name}"
    else:
        tensorboard_event_path = config.train.tensorboard_event_path
    tensorboard_event_path = Path(to_absolute_path(tensorboard_event_path))

    logger.info(f"TensorBoard event log path: {tensorboard_event_path}")
    tensorboard_writer = SummaryWriter(tensorboard_event_path)

    # Setup test result logging
    if config.train.test_log_dir:
        log_dir = Path(to_absolute_path(config.train.test_log_dir))
    else:
        log_dir = Path(to_absolute_path("logs")) / checkpoint_dir.name

    if not log_dir.exists():
        log_dir.mkdir(parents=True)
    log_path = log_dir / f"{checkpoint_dir.name}_test_log.json"

    # Init feature set
    feature_set = FeatureSet(
        Path(to_absolute_path(config.model.vocab_path)),
        feature_table_key=config.data.feature_table_key,
        feature_keys=config.data.input_keys,
    )

    # Init model
    model, criterions, optimizer, scheduler = init_model(
        tasks, config, feature_set, device, is_train=True
    )

    # Train
    train_model(
        model,
        criterions,
        optimizer,
        scheduler,
        dataloader,
        tasks,
        tensorboard_writer,
        checkpoint_dir,
        config,
        feature_set,
        num_epochs=num_epochs,
        device=device,
    )

    # Init test model
    _test_model = init_model(tasks, config, feature_set, device)
    logs = test_model(
        _test_model,
        checkpoint_dir,
        config.train.test_checkpoint_filename,
        tasks,
        dataloader,
        config,
        feature_set,
        tensorboard_writer=tensorboard_writer,
        device=device,
        logger=logger,
    )

    if config.train.save_test_log:
        # save test result log
        with open(log_path, "w") as file:
            json.dump(
                logs,
                file,
                ensure_ascii=False,
                indent=4,
                separators=(",", ": "),
            )

    tensorboard_writer.close()


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



from logging import getLogger
from pathlib import Path

import joblib
import numpy as np
from marine.data.feature.feature_table import (
    FEATURE_TABLES,
    PUNCTUATIONS,
    parse_accent_con_type,
)
from marine.utils.g2p_util import pron2mora

logger = getLogger(__name__)


class FeatureSet(object):
    """
    A converter for embedding features
    """

    def __init__(
        self,
        vocab_path,
        feature_table_key="unidic-csj",
        feature_keys=None,
        pad_token="[PAD]",
        unk_token="[UNK]",
    ):
        if isinstance(vocab_path, Path):
            self.vocab_path = vocab_path
        elif isinstance(vocab_path, str):
            self.vocab_path = Path(vocab_path)
        else:
            raise ValueError(
                f"Invalid type for vocab_path: {type(vocab_path)} (must be Path or str)"
            )

        self.pad_token = pad_token
        self.unk_token = unk_token
        self.default_tokens = [self.pad_token, self.unk_token]
        self.feature_table = FEATURE_TABLES[feature_table_key]

        if feature_keys:
            self.feature_keys = feature_keys
        else:
            self.feature_keys = list(self.feature_table.keys())

        self.feature_to_id = {key: [] for key in self.feature_keys}
        self.id_to_feature = {key: [] for key in self.feature_keys}

        self.init_feature_set()

    def _load_vocab(self):
        if not self.vocab_path.exists():
            logger.error(f"Vocab has not found : {self.vocab_path}")
            raise FileNotFoundError(f"Vocab has not found : {self.vocab_path}")

        vocab = joblib.load(self.vocab_path)
        logger.info(f"Vocab loaded from {self.vocab_path} : {len(vocab)} words")

        return vocab

    def init_feature_set(self):
        self._load_vocab()

        for key in self.feature_keys:
            if key == "surface":
                feature_set = self.default_tokens + self._load_vocab()
            else:
                if key not in self.feature_table.keys():
                    raise ValueError(
                        f"Feature key must be one of {self.feature_table.keys()}"
                    )
                feature_set = self.default_tokens + self.feature_table[key]

            feature_to_id = {
                feature_value: index for index, feature_value in enumerate(feature_set)
            }
            id_to_feature = {
                index: feature_value for feature_value, index in feature_to_id.items()
            }
            self.feature_to_id[key] = feature_to_id
            self.id_to_feature[key] = id_to_feature

    def convert_feature_to_id(self, feature_key, features):
        if feature_key not in self.feature_to_id:
            raise ValueError(
                f"Not initialized feature key: the key must be one of {self.feature_to_id}"
            )

        return np.array(
            [
                self.feature_to_id[feature_key].get(
                    value, self.feature_to_id[feature_key][self.unk_token]
                )
                for value in features
            ],
            dtype=np.uint8,
        )

    def convert_id_to_feature(self, feature_key, ids):
        if feature_key not in self.id_to_feature:
            raise ValueError(
                f"Not initialized feature key: the key must be one of {self.id_to_feature}"
            )

        return np.array(
            [
                self.id_to_feature[feature_key].get(value, self.unk_token)
                for value in ids
            ]
        )

    def convert_nodes_to_feature(self, nodes):
        features = {key: np.array([], np.uint8) for key in self.feature_to_id}

        # init morph boundary for inference
        features["morph_boundary"] = np.array([], np.uint8)

        for node in nodes:
            mora = self.convert_feature_to_id(
                "mora", pron2mora(node["pron"]) if node["pron"] else [node["surface"]]
            )
            morph_boundary = np.array([1] + ([0] * (len(mora) - 1)), dtype=np.uint8)

            # Push features
            features["mora"] = np.concatenate([features["mora"], mora], axis=0)
            features["morph_boundary"] = np.concatenate(
                [features["morph_boundary"], morph_boundary], axis=0
            )

            for key, table in self.feature_to_id.items():
                if key in ["mora", "morph_boundary"]:
                    continue

                if key == "a_con_type":
                    key = parse_accent_con_type(
                        node["accent_con_type"], node["pos"], unk_token=self.unk_token
                    )

                feature = table.get(node[key], table[self.unk_token])
                feature = np.array([feature] * len(mora), dtype=np.uint8)
                features[key] = np.concatenate([features[key], feature], axis=0)

        # First Mora could not be boundary
        # (boundary should be [0, 0, 1, 0, 0 ...])
        features["morph_boundary"][0] = 0

        return features

    def get_punctuation_ids(self):
        return [self.feature_to_id["mora"][punctuation] for punctuation in PUNCTUATIONS]



import re
import warnings
from pathlib import Path

import torch
from marine.data.feature.feature_set import FeatureSet
from marine.data.pad import Padsequence
from marine.models import (
    AttentionBasedLSTMDecoder,
    CRFDecoder,
    LinearDecoder,
    init_model,
)
from marine.utils.post_process import apply_postprocess_dict, load_postprocess_vocab
from marine.utils.pretrained import retrieve_pretrained_model
from marine.utils.util import (
    _convert_ap_based_accent_to_mora_based_accent,
    convert_open_jtalk_format_label,
    expand_word_label_to_mora,
    sequence_mask,
)
from omegaconf import OmegaConf
from pkg_resources import resource_filename
from torch.nn.utils.rnn import pad_sequence

BASE_DIR = Path(resource_filename("marine", ""))
DEFAULT_POSTPROCESS_VOCAB_DIR = BASE_DIR / "dict"


class Predictor:
    """Interface for inference of accent model."""

    def __init__(
        self,
        model_dir=None,
        version=None,
        postprocess_vocab_dir=None,
        device="cpu",
        skip_post_process=False,
    ):
        self.setup_model(model_dir, version, device)
        self.setup_postprocess_vocab(postprocess_vocab_dir, skip_post_process)

    def setup_model(self, model_dir, version, device):
        if model_dir is None:
            if version is None:
                self.model_dir = Path(retrieve_pretrained_model())
            else:
                self.model_dir = Path(retrieve_pretrained_model(version))
        elif isinstance(model_dir, str):
            self.model_dir = Path(model_dir)

        assert (
            isinstance(self.model_dir, Path) and self.model_dir.exists()
        ), f"Model directory doesn't exists: {self.model_dir.as_posix()}"

        self.device = device
        self.config = OmegaConf.load(self.model_dir / "config.yaml")
        self.tasks = self.config.data.output_keys

        if self.config.model.vocab_path is None:
            self.config.model.vocab_path = str(self.model_dir / "vocab.pkl")

        self.feature_set = FeatureSet(
            self.config.model.vocab_path,
            feature_table_key=self.config.data.feature_table_key,
            feature_keys=self.config.data.input_keys,
        )
        self.model = init_model(self.tasks, self.config, self.feature_set, self.device)
        self._load_states()

        self.collate_fn = Padsequence(
            self.config.data.input_keys,
            self.config.data.input_length_key,
            self.config.data.output_keys,
            self.config.data.output_sizes,
            is_inference=True,
        )

    def _load_states(self):
        states = torch.load(self.model_dir / "model.pth", map_location=self.device)
        self.model.load_state_dict(states["state_dict"])
        self.model.to(self.device)
        self.model.eval()

    def setup_postprocess_vocab(self, postprocess_vocab_dir, skip_post_process):
        # Setup for vocab for post-process
        if postprocess_vocab_dir is None:
            self.postprocess_vocab_dir = DEFAULT_POSTPROCESS_VOCAB_DIR

        elif isinstance(postprocess_vocab_dir, str):
            self.postprocess_vocab_dir = Path(postprocess_vocab_dir)

        assert (
            self.postprocess_vocab_dir.exists()
        ), f"Vocab directory doesn't exists: {self.postprocess_vocab_dir.as_posix()}"

        if skip_post_process:
            self.postprocess_vocab, self.epostprocess_targets = None, None
        else:
            self.postprocess_vocab = load_postprocess_vocab(
                self.postprocess_vocab_dir, self.tasks
            )
            self.postprocess_targets = {
                key: re.compile(f"({'|'.join(vocab.keys())})") if vocab else None
                for key, vocab in self.postprocess_vocab.items()
            }

    @torch.no_grad()
    def predict(
        self,
        sentences,
        accent_represent_mode="binary",
        annotates=None,
        require_open_jtalk_format=False,
    ):
        if accent_represent_mode not in ["binary", "high_low"]:
            raise NotImplementedError(
                (
                    f"Not supported representation mode {accent_represent_mode}:"
                    " Representation mode must be selected in binary and high_low"
                )
            )

        if require_open_jtalk_format and accent_represent_mode != "binary":
            warnings.warn(
                (
                    "If you want the format for OpenJTalk,",
                    "`accent_represent_mode` will be fixed as `binary`",
                )
            )
            accent_represent_mode = "binary"

        result = {}

        inputs, morph_boundary = self.extract_feature(sentences)
        result["mora"] = self.convert_to_mora(inputs)

        if annotates is not None:
            annotates = self.pad_annotate_label(
                annotates, mora=result["mora"], morph_boundary=morph_boundary
            )

        for task in self.tasks:
            if annotates is not None and task in annotates.keys():
                real_labels = self.convert_to_label(
                    task,
                    annotates[task],
                    inputs["mask"],
                    accent_represent_mode=accent_represent_mode,
                )
                prev_output = annotates[task]

            else:
                outputs = self.model(task, **inputs)

                if isinstance(self.model.decoders[task], CRFDecoder):
                    _, logits = outputs
                    ap_lengths, ap_outputs = None, None

                elif isinstance(self.model.decoders[task], LinearDecoder):
                    logits = outputs
                    ap_lengths, ap_outputs = None, None

                elif isinstance(self.model.decoders[task], AttentionBasedLSTMDecoder):
                    logits, _, ap_lengths = outputs
                    ap_outputs = inputs["prev_decoder_outputs"][
                        "accent_phrase_boundary"
                    ]

                real_labels = self.convert_to_label(
                    task,
                    logits,
                    inputs["mask"],
                    moras=result["mora"],
                    ap_lengths=ap_lengths,
                    ap_outputs=ap_outputs,
                    accent_represent_mode=accent_represent_mode,
                )

                if self.postprocess_vocab:
                    target = self.postprocess_targets[task]
                    vocab = self.postprocess_vocab[task]

                    if vocab:
                        for index, (nodes, label, mora, boundary) in enumerate(
                            zip(sentences, real_labels, result["mora"], morph_boundary)
                        ):
                            real_labels[index] = apply_postprocess_dict(
                                task,
                                nodes,
                                label,
                                mora,
                                boundary,
                                target,
                                vocab,
                                accent_represent_mode,
                            )

                prev_output = pad_sequence(
                    [torch.tensor(label) for label in real_labels], batch_first=True
                ).to(self.device)

            result[task] = real_labels
            inputs["prev_decoder_outputs"][task] = prev_output

        if require_open_jtalk_format:
            result = convert_open_jtalk_format_label(result, morph_boundary)

        return result

    def extract_feature(self, sentences):
        batch = [
            {
                "features": self.feature_set.convert_nodes_to_feature(nodes),
                "labels": None,
            }
            for nodes in sentences
        ]
        inputs, _, morph_boundary, _ = self.collate_fn(batch)

        embeddings = {
            key: inputs[key].to(self.device) for key in self.config.data.input_keys
        }

        inputs = {
            "embedding_features": embeddings,
            "lengths": inputs["mora_length"].cpu(),
            "mask": sequence_mask(inputs["mora_length"]).to(self.device),
            "prev_decoder_outputs": {},
        }

        return inputs, morph_boundary

    def convert_to_mora(self, inputs):
        return [
            list(self.feature_set.convert_id_to_feature("mora", mora[mask].tolist()))
            for mora, mask in zip(inputs["embedding_features"]["mora"], inputs["mask"])
        ]

    def convert_to_label(
        self,
        task,
        outputs,
        mora_masks,
        moras=None,
        ap_lengths=None,
        ap_outputs=None,
        accent_represent_mode="binary",
    ):
        predicts = []

        for index, (mora_mask, padded_predict) in enumerate(zip(mora_masks, outputs)):
            # for annoted label
            if len(padded_predict.shape) != 2:
                predict = padded_predict[mora_mask]
            else:
                if task == "accent_status" and ap_lengths is not None:
                    predict = padded_predict[: ap_lengths[index]]
                    ap_output = ap_outputs[index][mora_mask]

                    predict = torch.argmax(predict, dim=1)
                    predict = _convert_ap_based_accent_to_mora_based_accent(
                        predict,
                        ap_output,
                        mode=accent_represent_mode,
                        mora=moras[index],
                    )
                else:
                    predict = padded_predict[mora_mask]
                    predict = torch.argmax(predict, dim=1)
                    # convert 0-based label
                    predict = predict - 1

            predicts.append((predict).tolist())

        return predicts

    def pad_annotate_label(self, annotates, mora, morph_boundary):
        for key in annotates.keys():
            token_type = annotates[key]["token_type"]
            annotate = annotates[key]["labels"]

            if token_type == "morph":
                annotate = expand_word_label_to_mora(
                    annotate, mora, morph_boundary, key
                )
            elif token_type != "mora":
                raise ValueError(f"Token type must be morph or mora: {token_type}")

            if isinstance(annotate[0], list):
                annotate = [torch.tensor(ann) for ann in annotate]
            elif not isinstance(annotates[key][0], torch.Tensor):
                raise ValueError(
                    (
                        "Annoate labels must be List[List] or List[tensor]:"
                        f" {type(annotates[key][0])}"
                    )
                )

            annotates[key] = pad_sequence(annotate, batch_first=True).to(self.device)

        return annotates



from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class BiLSTMEncoder(nn.Module):
    def __init__(self, input_size, hidden_size, num_layers, shared_with=None):
        super().__init__()

        self.lstm = nn.LSTM(
            input_size,
            hidden_size,
            num_layers=num_layers,
            batch_first=True,
            bidirectional=True,
        )

    def forward(self, embeddings, lengths):
        # LSTM -> B * T * Hidden-size
        packed = pack_padded_sequence(
            embeddings, lengths, batch_first=True, enforce_sorted=True
        )
        logits, _ = self.lstm(packed)
        logits, _ = pad_packed_sequence(logits, batch_first=True)

        return logits



import json
from logging import getLogger
from pathlib import Path
from typing import Dict

import pytest
import torch
from marine.utils.metrics import MultiTaskMetrics, SentenceLevelAccuracy
from numpy.testing import assert_almost_equal
from pkg_resources import resource_filename

logger = getLogger("test")
BASE_DIR = Path(resource_filename("marine", ""))


@pytest.fixture
def sentence_level_accuracy() -> SentenceLevelAccuracy:
    return SentenceLevelAccuracy()


@pytest.fixture
def high_low_multi_task_metrics() -> MultiTaskMetrics:
    phase = "train"
    task_label_sizes = {"accent_phrase_boundary": 3}

    return MultiTaskMetrics(phase, task_label_sizes, device="cpu")


@pytest.fixture
def ap_multi_task_metrics() -> MultiTaskMetrics:
    phase = "train"
    task_label_sizes = {"accent_status": 21}
    require_ap_level_f1_score = True

    return MultiTaskMetrics(
        phase,
        task_label_sizes,
        require_ap_level_f1_score=require_ap_level_f1_score,
        device="cpu",
    )


@pytest.fixture
def full_multi_task_metrics() -> MultiTaskMetrics:
    phase = "train"
    task_label_sizes = {
        "intonation_phrase_boundary": 3,
        "accent_phrase_boundary": 3,
        "accent_status": 21,
    }
    require_ap_level_f1_score = True
    device = "cpu"

    return MultiTaskMetrics(
        phase,
        task_label_sizes,
        require_ap_level_f1_score=require_ap_level_f1_score,
        device=device,
    )


@pytest.fixture
def test_log_sample() -> Dict:
    logs = None
    sample_path = BASE_DIR.parent / "tests" / "samples" / "test_log_sample.json"
    with open(sample_path, "r") as file:
        logs = json.load(file)
    return logs


def test_sentence_level_accuracy(sentence_level_accuracy):
    for pred, target, mask, expect in [
        (
            torch.tensor([[1, 2, 1, 1, 2, 1, 0], [1, 2, 1, 1, 1, 0, 0]]),
            torch.tensor([[1, 1, 2, 1, 1, 2, 0], [1, 1, 2, 1, 1, 0, 0]]),
            torch.tensor(
                [
                    [True, True, True, True, True, True, False],
                    [True, True, True, True, True, False, False],
                ],
            ),
            0.0,
        ),
        (
            torch.tensor([[1, 1, 2, 1, 1, 2, 0], [1, 2, 1, 1, 1, 0, 0]]),
            torch.tensor([[1, 1, 2, 1, 1, 2, 0], [1, 1, 2, 1, 1, 0, 0]]),
            torch.tensor(
                [
                    [True, True, True, True, True, True, False],
                    [True, True, True, True, True, False, False],
                ],
            ),
            0.5,
        ),
        (
            torch.tensor([[1, 1, 2, 1, 1, 2, -1], [1, 2, 1, 1, 1, 0, 0]]),
            torch.tensor([[1, 1, 2, 1, 1, 2, -1], [1, 2, 1, 1, 1, 0, 0]]),
            torch.tensor(
                [
                    [True, True, True, True, True, True, False],
                    [True, True, True, True, True, False, False],
                ],
            ),
            1.0,
        ),
    ]:
        sentence_level_accuracy.update(pred, target, mask)
        score = sentence_level_accuracy.compute()
        assert_almost_equal(score, expect)
        sentence_level_accuracy.reset()


def test_ap_muti_task_metrics(ap_multi_task_metrics):
    for task, ap_pred, ap_target, ap_mask, kwargs, expect in [
        (
            "accent_status",
            # NOTE: AP-based accent label requires remove padding index (=0)
            # e,g, torch.tensor([4]) means 3rd mora has accent nucleus
            torch.tensor([[1, 1, 2], [1, 3, 0]]),
            torch.tensor([[1, 1, 3], [1, 2, 0]]),
            torch.tensor([[True, True, True], [True, True, False]]),
            {
                # NOTE: accent phrase boundary label requires remove padding index (=0)
                # e,g, torch.tensor([1, 2, 1]) means 2nd mora has accent phrase boundary
                "predicted_accent_phrase_boundaries": torch.tensor(
                    [[1, 2, 1, 1, 2, 1, 1, 0], [1, 2, 1, 1, 1, 0, 0, 0]]
                ),
                "target_accent_phrase_boundaries": torch.tensor(
                    [[1, 2, 1, 1, 2, 1, 1, 0], [1, 2, 1, 1, 1, 0, 0, 0]]
                ),
                "mora_seq_masks": torch.tensor(
                    [
                        [True, True, True, True, True, True, True, False],
                        [True, True, True, True, True, False, False, False],
                    ]
                ),
            },
            {
                "ap_level_f1_score": 0.3333333432674408,
                "mora_level_f1_score": 0.40000003576278687,
                "sentence_level_accuracy": 0.0,
            },
        ),
        (
            "accent_status",
            torch.tensor([[1, 1, 3], [1, 2, -1]]),
            torch.tensor([[1, 1, 3], [1, 3, -1]]),
            torch.tensor([[True, True, True], [True, True, False]]),
            {
                "predicted_accent_phrase_boundaries": torch.tensor(
                    [[1, 2, 1, 1, 2, 1, 1, 0], [1, 2, 1, 1, 1, 0, 0, 0]]
                ),
                "target_accent_phrase_boundaries": torch.tensor(
                    [[1, 2, 1, 1, 2, 1, 1, 0], [1, 2, 1, 1, 1, 0, 0, 0]]
                ),
                "mora_seq_masks": torch.tensor(
                    [
                        [True, True, True, True, True, True, True, False],
                        [True, True, True, True, True, False, False, False],
                    ]
                ),
            },
            {
                "ap_level_f1_score": 0.5555555820465088,
                "mora_level_f1_score": 0.699999988079071,
                "sentence_level_accuracy": 0.5,
            },
        ),
        (
            "accent_status",
            torch.tensor([[1, 1, 3], [1, 3, 0]]),
            torch.tensor([[1, 1, 3], [1, 3, 0]]),
            torch.tensor([[True, True, True], [True, True, False]]),
            {
                "predicted_accent_phrase_boundaries": torch.tensor(
                    [[1, 2, 1, 1, 2, 1, 1, 0], [1, 2, 1, 1, 1, 0, 0, 0]]
                ),
                "target_accent_phrase_boundaries": torch.tensor(
                    [[1, 2, 1, 1, 2, 1, 1, 0], [1, 2, 1, 1, 1, 0, 0, 0]]
                ),
                "mora_seq_masks": torch.tensor(
                    [
                        [True, True, True, True, True, True, True, False],
                        [True, True, True, True, True, False, False, False],
                    ]
                ),
            },
            {
                "ap_level_f1_score": 1.0,
                "mora_level_f1_score": 1.0,
                "sentence_level_accuracy": 1.0,
            },
        ),
    ]:
        ap_multi_task_metrics.update(task, ap_pred, ap_target, ap_mask, **kwargs)
        scores = ap_multi_task_metrics.compute()

        for score_name, expect_score in expect.items():
            assert_almost_equal(scores[task][score_name], expect_score)

        ap_multi_task_metrics.reset()


def test_ap_muti_task_metrics_by_log(test_log_sample, full_multi_task_metrics):
    """Unit test for MultiTaskMetrics by sample"""

    def _extract_label(key, log):
        label = [int(v) for v in log[key].split(",")]
        # (T) -> (1, T)
        label = torch.tensor(label).unsqueeze(0).to("cpu")
        return label

    expected_scores = {
        "intonation_phrase_boundary": {
            "mora_level_f1_score": 0.9472237825393677,
            "sentence_level_accuracy": 0.7064254283905029,
        },
        "accent_phrase_boundary": {
            "mora_level_f1_score": 0.993394672870636,
            "sentence_level_accuracy": 0.875554,
        },
        "accent_status": {
            "ap_level_f1_score": 0.6181702017784119,
            "mora_level_f1_score": 0.9536119699478149,
            "sentence_level_accuracy": 0.6894387006759644,
        },
    }

    for script_status in test_log_sample.values():
        for task_name, task_status in script_status.items():
            preds = _extract_label("predict", task_status)
            targets = _extract_label("target", task_status)

            if task_name == "accent_status":
                predicted_ap_boundary = _extract_label(
                    "predict", script_status["accent_phrase_boundary"]
                )
                target_ap_bondary = _extract_label(
                    "target", script_status["accent_phrase_boundary"]
                )
                # ap_mask: (1, T_ap)
                ap_seq_mask = torch.full(targets.shape, True).to("cpu")
                # mora_mask: (1, T_mora)
                mora_seq_mask = torch.full(target_ap_bondary.shape, True).to("cpu")

                full_multi_task_metrics.update(
                    task_name,
                    preds,
                    targets,
                    ap_seq_mask,
                    predicted_accent_phrase_boundaries=predicted_ap_boundary,
                    target_accent_phrase_boundaries=target_ap_bondary,
                    mora_seq_masks=mora_seq_mask,
                )
            else:
                masks = torch.full(targets.shape, True).to("cpu")
                full_multi_task_metrics.update(task_name, preds, targets, masks)

    scores = full_multi_task_metrics.compute()

    # verify keys the metric has is correct
    assert expected_scores.keys() == scores.keys()

    for task_name, task_status in expected_scores.items():
        for score_name, score in task_status.items():
            assert_almost_equal(score, scores[task_name][score_name])



import torch
from torch.nn.utils.rnn import pad_sequence


class Padsequence(object):
    def __init__(
        self,
        input_keys,
        input_length_key,
        output_keys,
        num_classes,
        is_inference=False,
        padding_idx=0,
    ):
        self.input_keys = input_keys
        self.input_length_key = input_length_key
        self.output_keys = output_keys
        self.num_classes = num_classes
        self.is_inference = is_inference
        self.padding_idx = padding_idx

    def pad_feature(self, inputs):
        padded_feature = {}

        for key in self.input_keys:
            feature = [
                torch.tensor(features[key], dtype=torch.int64) for features in inputs
            ]

            if key in self.input_length_key:
                padded_feature[f"{key}_length"] = torch.tensor(
                    [len(f) for f in feature], dtype=torch.int64
                )

            padded_x = pad_sequence(
                feature,
                batch_first=True,
                padding_value=self.padding_idx,
            )
            padded_feature[key] = padded_x

        return padded_feature

    def __call__(self, batch):
        # sort by length
        if not self.is_inference:
            batch = sorted(
                batch,
                key=lambda x: len(x["features"][self.input_length_key]),
                reverse=True,
            )

        inputs = [x["features"] for x in batch]
        padded_inputs = self.pad_feature(inputs)

        if not self.is_inference:
            padded_outputs = {
                key: {
                    "label": pad_sequence(
                        [
                            # Covnert 1-based label (for pad)
                            torch.tensor(x["labels"][key] + 1, dtype=torch.long)
                            for x in batch
                        ],
                        batch_first=True,
                        padding_value=self.padding_idx,
                    ),
                    "length": torch.tensor([len(x["labels"][key]) for x in batch]),
                }
                for key in self.output_keys
            }
            script_ids = [x["ids"] for x in batch]
        else:
            padded_outputs = None
            script_ids = None

        morph_boundary = [x["features"]["morph_boundary"] for x in batch]

        return padded_inputs, padded_outputs, morph_boundary, script_ids



import codecs
import re
from os.path import exists, join

from setuptools import find_packages, setup


def find_version(*file_paths: str) -> str:
    with codecs.open(join(*file_paths), "r") as fp:
        version_file = fp.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


if exists("README.md"):
    with open("README.md", "r") as fh:
        LONG_DESC = LONG_DESC = fh.read()
else:
    LONG_DESC = ""


setup(
    name="marine",
    version=find_version("marine", "__init__.py"),
    description="Marine: Multi-task learning based on Japanese accent estimation",
    packages=find_packages(),
    author="Byeongseon Park",
    author_email="6gsn.park@gmail.com",
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        "numpy >= 1.21.0",
        "torch >= 1.7.0",
        "hydra-core >= 1.1.0",
        "hydra_colorlog >= 1.1.0",
        "tqdm",
    ],
    extras_require={
        "dev": [
            "torchmetrics",
            "scikit-learn",
            "joblib",
            "docstr-coverage",
            "tensorboard",
            "matplotlib",
            "pytest",
            "pytest-cov",
            "docstr-coverage",
            "pysen",
            "black>=19.19b0,<=20.8",
            "flake8>=3.7,<4",
            "flake8-bugbear",
            "isort>=4.3,<5.2.0",
            "click<8.1.0",
        ],
        "docs": [
            "sphinx",
            "sphinx-autobuild",
            "sphinx_rtd_theme",
            "nbsphinx>=0.8.6",
            "Jinja2>=3.0.1",
            "pandoc",
            "ipython",
            "jupyter",
        ],
        "pyopenjtalk": ["pyopenjtalk"],
    },
    entry_points={
        "console_scripts": [
            "marine-jsut2corpus = marine.bin.jsut2corpus:entry",
            "marine-build-vocab = marine.bin.build_vocab:entry",
            "marine-pack-corpus = marine.bin.pack_corpus:entry",
            "marine-train = marine.bin.train:entry",
            "marine-test = marine.bin.test:entry",
        ],
    },
)



from __future__ import absolute_import, print_function, with_statement

import logging
import os
from os.path import dirname

format = "[%(asctime)s][%(name)s:%(module)s][%(levelname)s] - %(message)s"


def getLogger(verbose=0, filename=None, name="marine"):
    handlers = []

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(format))
    handlers.append(stream_handler)

    if filename is not None:
        os.makedirs(dirname(filename), exist_ok=True)
        file_handler = logging.FileHandler(filename=filename)
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(logging.Formatter(format))
        handlers.append(file_handler)

    logging.basicConfig(handlers=handlers)

    logger = logging.getLogger(name)
    if verbose >= 100:
        logger.setLevel(logging.DEBUG)
    elif verbose > 0:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARN)

    return logger



import argparse
import sys
from collections import defaultdict
from pathlib import Path

from joblib import dump
from marine.logger import getLogger
from marine.utils.util import load_json_corpus
from tqdm import tqdm

logger = None


def get_parser():
    parser = argparse.ArgumentParser(
        description="Generate vocabulary file for word-embedding",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("in_path", type=Path, help="Path or directory for feature file")
    parser.add_argument("out_dir", type=Path, help="Output directory")
    parser.add_argument(
        "--min_freq",
        "-m",
        type=int,
        default=2,
        help="""
        Minimum word frequency in the whole corpus,
        which to judge whether the word include in the vocabulary
        """,
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Whether print log for debug",
    )
    return parser


def count_words(words):
    freqs = defaultdict(int)

    for word in tqdm(words, "Counting word", leave=False):
        freqs[word] += 1

    freqs = sorted(freqs.items(), key=lambda x: x[1], reverse=True)

    return freqs


def filter_words(freqs, min_freq):
    return list(filter(lambda x: x[1] >= min_freq, freqs))


def save_vocab(freqs, output_dir):
    words = [surface for surface, _ in freqs]
    dump(words, output_dir / "vocab.pkl", compress=True)


def entry(argv=sys.argv):
    global logger

    args = get_parser().parse_args(argv[1:])
    logger = getLogger(args.verbose)
    logger.debug(f"Loaded parameters: {args}")

    input_path = args.in_path
    output_dir = args.out_dir

    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    corpus = load_json_corpus(input_path)
    words = [node["surface"] for script in corpus for node in script["nodes"]]
    logger.info(f"Loaded {len(words):,} words")

    freqs = count_words(words)
    logger.info(f"Get {len(freqs):,} unique words")

    freqs = filter_words(freqs, args.min_freq)
    logger.info(f"Filtered {len(freqs):,} words")

    save_vocab(freqs, output_dir)


if __name__ == "__main__":
    sys.exit(entry())



from logging import getLogger

from marine.bin.jsut2corpus import parse_jsut_annotation

logger = getLogger("test")


def test_jsut_parser():
    for (
        jsut_annotaion,
        accent_status_seq_level,
        accent_status_represent_mode,
        expect,
    ) in [
        (
            "^ム]シロ#ロ[ンゲノホ]ーガ#ハ[ゲヤス]イッテ#キ[ータゾ?$",
            "ap",
            "binary",
            {
                "pron": "ムシロロンゲノホーガハゲヤスイッテキータゾ",
                "accent_status": "1,5,4,0",
                "accent_phrase_boundary": "0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0",
                "intonation_phrase_boundary": "0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0",
            },
        ),
        (
            "^ム]シロ#ロ[ンゲノホ]ーガ#ハ[ゲヤス]イッテ#キ[ータゾ?$",
            "mora",
            "binary",
            {
                "pron": "ムシロロンゲノホーガハゲヤスイッテキータゾ",
                "accent_status": "1,0,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0",
                "accent_phrase_boundary": "0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0",
                "intonation_phrase_boundary": "0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0",
            },
        ),
        (
            "^ム]シロ#ロ[ンゲノホ]ーガ#ハ[ゲヤス]イッテ#キ[ータゾ?$",
            "mora",
            "high_low",
            {
                "pron": "ムシロロンゲノホーガハゲヤスイッテキータゾ",
                "accent_status": "1,0,0,0,1,1,1,1,0,0,0,1,1,1,0,0,0,0,1,1,1",
                "accent_phrase_boundary": "0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0",
                "intonation_phrase_boundary": "0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0",
            },
        ),
    ]:
        result = parse_jsut_annotation(
            jsut_annotaion, accent_status_seq_level, accent_status_represent_mode
        )
        assert result == expect



# Acknowledgement: some of the code was adapted from ttslearn
#  Copyright 2021 Ryuichi Yamamoto (MIT License)

import os
import shutil
import tarfile
from os.path import join
from pathlib import Path
from urllib.request import urlretrieve

from tqdm.auto import tqdm

DEFAULT_CACHE_DIR = join(os.path.expanduser("~"), ".cache", "marine")
CACHE_DIR = os.environ.get("MARINE_CACHE_DIR", DEFAULT_CACHE_DIR)

DEFAULT_VERSION = "v0.0.2"
MODEL_BASE_URL = "https://github.com/6gsn/marine/releases/download/"


# https://github.com/tqdm/tqdm#hooks-and-callbacks
class _TqdmUpTo(tqdm):  # type: ignore
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)


def retrieve_pretrained_model(version=None):
    """Retrieve pretrained model from local cache or download from GitHub.
    Args:
        version (str): Version of pretrained model.
    Returns:
        str: Path to the pretrained model.
    Raises:
        ValueError: If the pretrained model is not found.
    Examples:
        >>> from marine.utils.pretrained import retrieve_pretrained_model
        >>> from marine.predict import Predictor
        >>> model_dir = retrieve_pretrained_model("v0.0.2")
        >>> predictor = Tacotron2PWGTTS(model_dir=model_dir, device="cpu")
    """

    if version is None:
        version = DEFAULT_VERSION
    elif not isinstance(version, str):
        raise TypeError(f"version must be str not {type(version)}")

    url = MODEL_BASE_URL + f"{version}/model.tar.gz"

    # NOTE: assuming that filename and extracted is the same
    out_dir = Path(CACHE_DIR) / version
    filename = Path(CACHE_DIR) / f"{version}/model.tar.gz"

    # re-download models
    if out_dir.exists() and len(list(out_dir.glob("*.pth"))) == 0:
        shutil.rmtree(out_dir)

    if not out_dir.exists():
        print(
            "The use of pre-trained models is permitted for non-commercial use only."
            "Please visit https://github.com/6gsn/marine to confirm the license."
        )
        print('Downloading: "{}"'.format(url))

        out_dir.mkdir(parents=True, exist_ok=True)

        with _TqdmUpTo(
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            miniters=1,
            desc=f"{version}/model.tar.gz",
        ) as t:  # all optional kwargs
            urlretrieve(url, filename, reporthook=t.update_to)
            t.total = t.n
        with tarfile.open(filename, mode="r|gz") as f:
            f.extractall(path=out_dir)
        os.remove(filename)

    return out_dir



import torch
from marine.utils.util import convert_ap_based_accent_to_mora_based_accent
from torchmetrics import F1Score, Metric, MetricCollection


class SentenceLevelAccuracy(Metric):
    """Metrics to calculate sentence level accuray."""

    full_state_update: bool = True

    def __init__(self):
        super().__init__()
        self.add_state("correct", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: torch.Tensor, targets: torch.Tensor, masks: torch.Tensor):
        """Update variables for accuracy."""

        # preds: (B, T), target: (B, T), mask: (B, T)
        # sequence_level_matchs: (B)
        sequence_level_matchs = torch.LongTensor(
            [
                (target[mask] == pred[mask]).all()
                for pred, target, mask in zip(preds, targets, masks)
            ]
        )

        self.correct += torch.sum(sequence_level_matchs)
        self.total += sequence_level_matchs.numel()  # == batch size

    def compute(self):
        """Compute accuracy using variables."""
        return self.correct.float() / self.total


class MultiTaskMetrics(object):
    """Metrics to calculate scores."""

    def __init__(
        self,
        phase,
        task_label_sizes,
        average="macro",
        accent_represent_mode="binary",
        require_ap_level_f1_score=False,
        device="cpu",
    ):
        self.phase = phase
        self.tasks = task_label_sizes.keys()
        self.accent_represent_mode = accent_represent_mode
        self.require_ap_level_f1_score = require_ap_level_f1_score
        self.device = device

        self.metrics = {
            task_name: MetricCollection(
                {
                    "mora_level_f1_score": F1Score(
                        num_classes=2,  # the AN label represents High/Low or non-AN/AN
                        average=average,
                    ).to(device),
                    "ap_level_f1_score": F1Score(
                        num_classes=task_label_size, average=average
                    ).to(device),
                    "sentence_level_accuracy": SentenceLevelAccuracy().to(device),
                }
                if task_name == "accent_status" and require_ap_level_f1_score
                else {
                    "mora_level_f1_score": F1Score(
                        num_classes=task_label_size, average=average
                    ).to(device),
                    "sentence_level_accuracy": SentenceLevelAccuracy().to(device),
                }
            ).to(device)
            for task_name, task_label_size in task_label_sizes.items()
        }

    def update(self, task, preds, targets, masks, padding_idx=0, **kwargs):
        """Update variables for accuracy."""
        assert task in self.tasks, f"Not initialized task: {task} not in {self.tasks}"

        masked_preds, masked_targets = preds[masks], targets[masks]

        # Verify that masked target label sequence not includes [PAD] token
        # TODO: This assert could be omited using `ignore_index` of `F1Score`
        # However, the option dosen't behave as explained until 2022/08/10
        # see https://github.com/Lightning-AI/metrics/issues/613
        assert padding_idx not in masked_targets

        if task == "accent_status" and self.require_ap_level_f1_score:
            mora_pred, mora_target = self._convert_ap_seq_to_mora_seq(
                preds, targets, masks, **kwargs
            )
            self.metrics[task]["mora_level_f1_score"].update(mora_pred, mora_target)
            self.metrics[task]["ap_level_f1_score"].update(masked_preds, masked_targets)
        else:
            self.metrics[task]["mora_level_f1_score"].update(
                masked_preds, masked_targets
            )

        self.metrics[task]["sentence_level_accuracy"].update(preds, targets, masks)

    def compute(self):
        """Compute accuracy using variables."""
        return {
            task_name: {
                score_name: metrics.compute().cpu().item()
                for score_name, metrics in self.metrics[task_name].items()
            }
            for task_name in self.tasks
        }

    def reset(self):
        """Reset all variables in metrics."""
        for task_name in self.tasks:
            for metrics in self.metrics[task_name].values():
                metrics.reset()

    def _convert_ap_seq_to_mora_seq(
        self,
        preds,
        targets,
        ap_seq_masks,
        predicted_accent_phrase_boundaries,
        target_accent_phrase_boundaries,
        mora_seq_masks,
    ):
        """Convert accent phrase-based accent status sequence to mora-based."""
        mora_preds = convert_ap_based_accent_to_mora_based_accent(
            preds,
            predicted_accent_phrase_boundaries,
            ap_seq_masks,
            mora_seq_masks,
            self.accent_represent_mode,
        )
        mora_targets = convert_ap_based_accent_to_mora_based_accent(
            targets,
            target_accent_phrase_boundaries,
            ap_seq_masks,
            mora_seq_masks,
            self.accent_represent_mode,
        )

        return (
            torch.LongTensor(mora_preds).to(self.device),
            torch.LongTensor(mora_targets).to(self.device),
        )



from logging import getLogger
from pathlib import Path

from hydra.utils import to_absolute_path
from joblib import load
from torch.utils.data import DataLoader

from .dataset import AccentDataset
from .pad import Padsequence

logger = getLogger(__name__)


def load_dataset(config, phases=None):
    dataloader = {}

    data_dir = Path(to_absolute_path(config.data.data_dir))

    if phases is None:
        phases = ["train", "val", "test"]
    elif not isinstance(phases, list):
        raise TypeError(f"Unvailable values: {phases}")

    for phase in phases:
        is_train = phase == "train"
        targets = ["features", "labels", "ids"]
        data = {}

        for target in targets:
            data_path = data_dir / phase / f"{target}.pkl"
            data[target] = load(data_path)

        dataset = AccentDataset(data)

        if logger is not None:
            logger.info(f"{phase} data size : {len(dataset):,}")

        if is_train:
            shuffle = True
        else:
            shuffle = False

        dataloader[phase] = DataLoader(
            dataset,
            batch_size=config.data.batch_size,
            shuffle=shuffle,
            collate_fn=Padsequence(
                input_keys=config.data.input_keys,
                input_length_key=config.data.input_length_key,
                output_keys=config.data.output_keys,
                num_classes=config.data.output_sizes,
            ),
            num_workers=config.data.num_workers,
        )

    return dataloader



from .util import CONNECTABLE_MORA, HALF_PUNCTUATION


def represent_syllable_boundary(index, moras, len_phonemes):
    """
    Represent syllable boundary by 2 types
    - types:
        - 0: non-boundary
        - 1: syllable boundary
    """

    if moras[index] in HALF_PUNCTUATION:
        return [1]

    # Init the boundary into normal boundary
    # e.g., か -> [k, a] -> [0, 1]
    #       わ -> [wa] -> [1]
    boundary = [0] * (len_phonemes - 1) + [1]

    # remove syllable boundaray if next mora is connectable mora (i.e. ッ, ン)
    # e.g., [ ... "デ", "ン" ] -> [ ... [d0, e1], [N1] ] -> [ ... [d0, e0], [N1] ]
    if index + 1 < len(moras):
        if (
            moras[index] not in CONNECTABLE_MORA
            and moras[index + 1] in CONNECTABLE_MORA
        ):
            boundary[-1] = 0
    else:
        boundary[-1] = 1

    return boundary
